unit SquareCA;

interface

uses
  Classes, Windows, GR32, SysUtils, Dialogs;

const
  Border = 4;
  NoNeighbors = 8;
  NoColors = 16;

type
  TCell =
  record
    State: Integer;
    NewState: Integer;
  end;
  TSelected = (None, TopLeftMarked, GroupSelected, Removing);

var
  DisplayWidth, DisplayHeight: Integer;
  Bitmap: TBitmap32;
  GridSize: Integer;
  NoRows, NoColumns: Integer;
  LeftBorder, TopBorder: Integer;
  Cells: array of TCell;
  SavedCells: array of Integer;
  GridLocation: TPoint;
  Neighbors: array[0..NoNeighbors - 1] of Integer;
  Colors: array[0..NoColors - 1] of TColor32;
  GridColor, SelectorColor: TColor32;
  SelectedState: Integer;
  Selected: TSelected;
  SelectedTopLeft, SelectedBottomRight: TPoint;
  X1, Y1, XN, YN: Integer;
  SX1, SY1, SXN, SYN: Integer;
  SelectionMoving: Boolean;
  MX, MY: Integer;
  Clear: Boolean;
  Period: Integer;

procedure InitializeCA;
procedure CreateBitmap;
procedure SetSize;
procedure CreateCells;
procedure DestroyCells;
procedure GetGridLocation(I: Integer);
procedure GetNeighbors(I: Integer);
procedure InvalidateStates;
procedure DrawCell(Cell: TCell; X, Y: Integer);
procedure DrawCells;
procedure ClearInside;
procedure ClearOutside;
procedure CopyNewStates;
procedure FlipX;
procedure FlipY;
procedure Rotate;
procedure RotateClockwise;
procedure RotateAntiClockwise;
procedure SelectAll;
procedure SelectSquare;
procedure Center;
procedure RandomSelection;
procedure SaveSelection;
procedure RestoreSelection;
procedure SaveState;
function CompareState: Boolean;
procedure StartMovingSelectedCells(Shift: TShiftState; X, Y: Integer);
procedure MoveSelectedCells(X, Y: Integer);
procedure StopMovingSelectedCells(Shift: TShiftState; X, Y: Integer);
procedure ClearCells;
procedure RandomCells;
procedure ResetNewStates;
procedure Step;
procedure ProcessMouseClick(X, Y: Integer; Shift: TShiftState);
procedure LoadPattern(FileName: String);
procedure SavePattern(FileName: String);

implementation

uses
  Display, Options, GeneralBinary, WeightedGenerations, InteractingStates,
  AnalysisResult;

procedure InitializeCA;
begin
  GridSize := 10;
  Selected := None;
  SelectionMoving := False;
  CreateBitmap;
  SetSize;
end;

procedure CreateBitmap;
begin
  Bitmap.Free;
  Bitmap := TBitmap32.Create;
  Bitmap.Width := DisplayWidth;
  Bitmap.Height := DisplayHeight;
  Bitmap.PenColor := clBlack32;
  Bitmap.FillRect(0, 0, Bitmap.Width, Bitmap.Height, clWhite32);
end;

procedure SetSize;
begin
  NoColumns := (DisplayWidth - 2 * Border) div GridSize;
  NoRows := (DisplayHeight - 2 * Border) div GridSize;
  LeftBorder := (DisplayWidth - NoColumns * GridSize) div 2;
  TopBorder := (DisplayHeight - NoRows * GridSize) div 2;
  DisplayForm.Statusbar.Panels[1].Text :=
    IntToStr(NoColumns) + 'x' + IntToStr(NoRows);
end;

procedure CreateCells;
var
  I: Integer;
begin
  SetLength(Cells, NoRows * NoColumns);
  for I := 0 to Length(Cells) - 1 do
  begin
    Cells[I].State := -1;
    Cells[I].NewState := 0;
  end;
end;

procedure DestroyCells;
begin
  Cells := nil;
end;

procedure GetGridLocation(I: Integer);
begin
  GridLocation.X := LeftBorder + (I mod NoColumns) * GridSize;
  GridLocation.Y := TopBorder + (I div NoColumns) * GridSize;
end;

procedure GetNeighbors(I: Integer);
var
  X, Y: Integer;

procedure AddNeighbor(I, X, Y: Integer);
begin
  Neighbors[I] :=
    ((NoRows + Y) mod NoRows) * NoColumns + (NoColumns + X) mod NoColumns;
end;

begin
  X := I mod NoColumns;
  Y := I div NoColumns;
  AddNeighbor(0, X - 1, Y - 1);
  AddNeighbor(1, X, Y - 1);
  AddNeighbor(2, X + 1, Y - 1);
  AddNeighbor(3, X - 1, Y);
  AddNeighbor(4, X + 1, Y);
  AddNeighbor(5, X - 1, Y + 1);
  AddNeighbor(6, X, Y + 1);
  AddNeighbor(7, X + 1, Y + 1);
end;

procedure DrawCell(Cell: TCell; X, Y: Integer);
var
  BrushColor: TColor32;
  XN, YN: Integer;
begin
  with Bitmap do
  begin
    BrushColor := Colors[Cell.State];
    if GridSize < 4 then
      PenColor := BrushColor;
    if GridSize > 3 then
    begin
      XN := X + GridSize + 1;
      YN := Y + GridSize + 1;
      FrameRectS(X, Y, XN, YN, PenColor);
      FillRectS(X + 1, Y + 1, XN - 1, YN - 1, BrushColor);
    end
    else if GridSize > 1 then
    begin
      XN := X + GridSize + 1;
      YN := Y + GridSize + 1;
      FillRectS(X, Y, XN - 1, YN - 1, BrushColor);
    end
    else
      Pixels[X, Y] := BrushColor;
  end;
end;

procedure InvalidateStates;
var
  I: Integer;
begin
  for I := 0 to Length(Cells) - 1 do
    Cells[I].State := -1;
end;

procedure DrawCells;
var
  I: Integer;
  X, Y: Integer;
begin
  if Selected = None then
  begin
    Bitmap.PenColor := GridColor;
    for I := 0 to Length(Cells) - 1 do
    begin
      if Cells[I].NewState <> Cells[I].State then
      begin
        Cells[I].State := Cells[I].NewState;
        GetGridLocation(I);
        DrawCell(Cells[I], GridLocation.X, GridLocation.Y);
      end;
    end;
  end
  else
  begin
    for I := 0 to Length(Cells) - 1 do
    begin
      GetGridLocation(I);
      X := GridLocation.X;
      Y := GridLocation.Y;
      if Selected = Removing then
        Bitmap.PenColor := GridColor
      else if (X >= SelectedTopLeft.X) and (X <= SelectedBottomRight.X) and
              (Y >= SelectedTopLeft.Y) and (Y <= SelectedBottomRight.Y) then
        Bitmap.PenColor := SelectorColor
      else
        Bitmap.PenColor := GridColor;
      DrawCell(Cells[I], X, Y);
    end;
    X := SelectedBottomRight.X + GridSize;
    Y := SelectedBottomRight.Y + GridSize;
    if Selected = Removing then
      Selected := None
    else
    begin
      Bitmap.PenColor := SelectorColor;
      with Bitmap do
      begin
        MoveTo(SelectedTopLeft.X, SelectedTopLeft.Y);
        LineToS(X, SelectedTopLeft.Y);
        LineToS(X, Y);
        LineToS(SelectedTopLeft.X, Y);
        LineToS(SelectedTopLeft.X, SelectedTopLeft.Y);
      end;
    end;
  end;
  DisplayForm.Image.Bitmap.Draw(0, 0, Bitmap);
end;

procedure ClearInside;
var
  X, Y: Integer;
begin
  for X := X1 to XN do
    for Y := Y1 to YN do
      Cells[Y * NoColumns + X].State := 0;
end;

procedure ClearOutside;
var
  I: Integer;
  X, Y: Integer;
begin
  for I := 0 to Length(Cells) - 1 do
  begin
    GetGridLocation(I);
    X := GridLocation.X;
    Y := GridLocation.Y;
    if (X < SelectedTopLeft.X) or (X > SelectedBottomRight.X) or
       (Y < SelectedTopLeft.Y) or (Y > SelectedBottomRight.Y) then
      Cells[I].State := 0;
  end;
end;

procedure CopyNewStates;
var
  X, Y: Integer;
  I: Integer;
begin
  for X := X1 to XN do
    for Y := Y1 to YN do
    begin
      I := Y * NoColumns + X;
      Cells[I].State := Cells[I].NewState;
    end;
end;

procedure FlipX;
var
  X, Y, YJ: Integer;
  I, J: Integer;
begin
  for X := X1 to XN do
  begin
    YJ := YN;
    for Y := Y1 to YN do
    begin
      I := Y * NoColumns + X;
      J := YJ * NoColumns + X;
      Cells[J].NewState := Cells[I].State;
      Dec(YJ);
    end;
  end;
  CopyNewStates;
end;

procedure FlipY;
var
  X, Y, XJ: Integer;
  I, J: Integer;
begin
  for Y := Y1 to YN do
  begin
    XJ := XN;
    for X := X1 to XN do
    begin
      I := Y * NoColumns + X;
      J := Y * NoColumns + XJ;
      Cells[J].NewState := Cells[I].State;
      Dec(XJ);
    end;
  end;
  CopyNewStates;
end;

procedure Rotate;
var
  X, Y: Integer;
  I, J: Integer;
begin
  X := XN - X1;
  Y := YN - Y1;
  if X <> Y then
  begin
    if X > Y then
      XN := X1 + Y
    else
      YN := Y1 + X;
    I := YN * NoColumns + XN;
    GetGridLocation(I);
    SelectedBottomRight := GridLocation;
  end;
  for Y := Y1 to YN do
    for X := X1 to XN do
    begin
      I := Y * NoColumns + X;
      J := (X - X1 + Y1) * NoColumns + Y - Y1 + X1;
      Cells[J].NewState := Cells[I].State;
    end;
  CopyNewStates;
end;

procedure RotateClockwise;
begin
  Rotate;
  FlipY;
end;

procedure RotateAntiClockwise;
begin
  Rotate;
  FlipX;
end;

procedure SelectAll;
var
  I: Integer;
  X, Y: Integer;
begin
  X1 := NoColumns - 1;
  Y1 := NoRows - 1;
  XN := 0;
  YN := 0;
  Clear := True;
  for I := 0 to Length(Cells) - 1 do
  begin
    if Cells[I].State > 0 then
    begin
      Clear := False;
      X := I mod NoColumns;
      if X < X1 then
        X1 := X;
      if X > XN then
        XN := X;
      Y := I div NoColumns;
      if Y < Y1 then
        Y1 := Y;
      if Y > YN then
        YN := Y;
    end;
  end;
  if Clear then
    Exit;
  I := Y1 * NoColumns + X1;
  GetGridLocation(I);
  SelectedTopLeft := GridLocation;
  I := YN * NoColumns + XN;
  GetGridLocation(I);
  SelectedBottomRight := GridLocation;
  Selected := GroupSelected;
end;

procedure SelectSquare;
var
  D: Integer;
  I: Integer;
begin
  SelectAll;
  if Clear then
    Exit;
  D := XN - X1 - YN + Y1;
  if D > 0 then
  begin
    Inc(YN, D);
    if YN > (NoRows - 1) then
    begin
      Dec(Y1, YN - NoRows + 1);
      YN := NoRows - 1;
      if Y1 < 0 then
        Y1 := 0;
    end;
  end
  else
  begin
    Inc(XN, -D);
    if XN > (NoColumns - 1) then
    begin
      Dec(X1, XN - NoColumns + 1);
      XN := NoColumns - 1;
      if X1 < 0 then
        X1 := 0;
    end;
  end;
  I := Y1 * NoColumns + X1;
  GetGridLocation(I);
  SelectedTopLeft := GridLocation;
  I := YN * NoColumns + XN;
  GetGridLocation(I);
  SelectedBottomRight := GridLocation;
end;

procedure Center;
var
  X, Y: Integer;
  DX, DY: Integer;
  I, J: Integer;
begin
  ClearCells;
  DX := ((NoColumns - XN + X1) div 2) - X1;
  DY := ((NoRows - YN + Y1) div 2) - Y1;
  for Y := Y1 to YN do
    for X := X1 to XN do
    begin
      I := Y * NoColumns + X;
      J := (Y + DY) * NoColumns + X + DX;
      Cells[J].NewState := Cells[I].State;
    end;
end;

procedure RandomSelection;
var
  Density, D: Integer;
  X, Y: Integer;
begin
  with OptionsDialog.RandomDensitySelector do
    Density := StrToInt(Items[ItemIndex]);
  Randomize;
  for Y := Y1 to YN do
    for X := X1 to XN do
    begin
      Cells[Y * NoColumns + X].NewState := 0;
      D := Random(100);
      if D < Density then
        Cells[Y * NoColumns + X].NewState := Random(MaxState + 1);
    end;
  CopyNewStates;
end;

procedure SaveSelection;
var
  X, Y: Integer;
  I: Integer;
begin
  I := 0;
  for Y := Y1 to YN do
    for X := X1 to XN do
    begin
      Cells[I].NewState := Cells[Y * NoColumns + X].State;
      Inc(I);
    end;
end;

procedure RestoreSelection;
var
  X, Y: Integer;
  I: Integer;
begin
  I := 0;
  for Y := Y1 to YN do
    for X := X1 to XN do
    begin
      if Cells[I].NewState > 0 then
        Cells[Y * NoColumns + X].State := Cells[I].NewState;
      Inc(I);
    end;
  for I := 0 to Length(Cells) - 1 do
    Cells[I].NewState := Cells[I].State;
end;

procedure SaveState;
var
  I: Integer;
  X, Y: Integer;
begin
  SX1 := NoColumns - 1;
  SY1 := NoRows - 1;
  SXN := 0;
  SYN := 0;
  for I := 0 to Length(Cells) - 1 do
  begin
    if Cells[I].State > 0 then
    begin
      X := I mod NoColumns;
      if X < SX1 then
        SX1 := X;
      if X > SXN then
        SXN := X;
      Y := I div NoColumns;
      if Y < SY1 then
        SY1 := Y;
      if Y > SYN then
        SYN := Y;
    end;
  end;
  SetLength(SavedCells, (SXN - SX1 + 1) * (SYN - SY1 + 1));
  I := 0;
  for Y := SY1 to SYN do
    for X := SX1 to SXN do
    begin
      SavedCells[I] := Cells[Y * NoColumns + X].State;
      Inc(I);
    end;
  Period := 0;
end;

function CompareState: Boolean;
var
  I: Integer;
  X, Y: Integer;

function MakeString(S1, S2: String): String;
const
  Blanks = '                        ';
var
  NoBlanks: Integer;
begin
  NoBlanks := 24 - Length(S1) - Length(S2);
  Result := S1 + Copy(Blanks, 1, NoBlanks) + S2;
end;

begin
  X1 := NoColumns - 1;
  Y1 := NoRows - 1;
  XN := 0;
  YN := 0;
  for I := 0 to Length(Cells) - 1 do
  begin
    if Cells[I].NewState > 0 then
    begin
      X := I mod NoColumns;
      if X < X1 then
        X1 := X;
      if X > XN then
        XN := X;
      Y := I div NoColumns;
      if Y < Y1 then
        Y1 := Y;
      if Y > YN then
        YN := Y;
    end;
  end;
  Inc(Period);
  Result := False;
  if ((XN - X1) <> (SXN - SX1)) or ((YN - Y1) <> (SYN - SY1)) then
  begin
    for I := 0 to Length(Cells) - 1 do
      Cells[I].State := Cells[I].NewState;
    Exit;
  end;
  I := 0;
  for Y := Y1 to YN do
    for X := X1 to XN do
    begin
      if SavedCells[I] <> Cells[Y * NoColumns + X].NewState then
      begin
        for I := 0 to Length(Cells) - 1 do
          Cells[I].State := Cells[I].NewState;
        Exit;
      end;
      Inc(I);
    end;
  with AnalysisResultForm do
  begin
    if ((X1 - SX1) = 0) and ((Y1 - SY1) = 0) then
      if Period = 1 then
        ObjectType.Caption := 'Object is Stable'
      else
        ObjectType.Caption := 'Object is an Oscillator'
    else
      ObjectType.Caption := 'Object is a Ship';
    PeriodText.Caption := MakeString('Period', IntToStr(Period));
    XDisplacement.Caption := MakeString('X Displacement', IntToStr(X1 - SX1));
    YDisplacement.Caption := MakeString('Y Displacement', IntToStr(Y1 - SY1));
    ShowModal;
  end;
  Result := True;
end;

procedure DrawSelection;
var
  C: TCell;
  X, Y: Integer;
  I: Integer;
begin
  I := 0;
  for Y := Y1 to YN do
    for X := X1 to XN do
    begin
      C.State := Cells[I].NewState;
      Bitmap.PenColor := SelectorColor;
      if C.State > 0 then
        DrawCell(C, LeftBorder + X * GridSize, TopBorder + Y * GridSize);
      Inc(I);
    end;
end;

procedure StartMovingSelectedCells(Shift: TShiftState; X, Y: Integer);
begin
  SaveSelection;
  SelectionMoving := True;
  MX:= X;
  MY:= Y;
end;

procedure MoveSelectedCells(X, Y: Integer);
var
  DX, DY: Integer;

procedure FitInDisplay;
begin
  if X1 < 0 then
  begin
    Inc(XN, -X1);
    X1 := 0;
  end
  else if XN >= NoColumns then
  begin
    Dec(X1, XN - NoColumns + 1);
    XN := NoColumns - 1;
  end;
  if Y1 < 0 then
  begin
    Inc(YN, -Y1);
    Y1 := 0;
  end
  else if YN >= NoRows then
  begin
    Dec(Y1, YN - NoRows + 1);
    YN := NoRows - 1;
  end;
end;

begin
  DX := (X - MX) div GridSize;
  DY := (Y - MY) div GridSize;
  if (abs(DX) >= 1) or (abs(DY) >= 1) then
  begin
    Inc(X1, DX);
    Inc(Y1, DY);
    Inc(XN, DX);
    Inc(YN, DY);
    FitInDisplay;
    SelectedTopLeft.X := LeftBorder + X1 * GridSize;
    SelectedTopLeft.Y := TopBorder + Y1 * GridSize;
    SelectedBottomRight.X := LeftBorder + XN * GridSize;
    SelectedBottomRight.Y := TopBorder + YN * GridSize;
    DrawCells;
    DisplayForm.Refresh;
    DrawSelection;
    DisplayForm.Image.Bitmap.Draw(0, 0, Bitmap);
    MX := X;
    MY := Y;
  end
end;

procedure StopMovingSelectedCells(Shift: TShiftState; X, Y: Integer);
begin
  RestoreSelection;
  SelectionMoving := False;
end;

procedure ClearCells;
var
  I: Integer;
begin
  for I := 0 to Length(Cells) - 1 do
    Cells[I].NewState := 0;
end;

procedure RandomCells;
var
  Density: Integer;
  D, I: Integer;
begin
  with OptionsDialog.RandomDensitySelector do
    Density := StrToInt(Items[ItemIndex]);
  ClearCells;
  Randomize;
  for I := 0 to Length(Cells) - 1 do
  begin
    D := Random(100);
    if D < Density then
      Cells[I].NewState := Random(MaxState + 1);
  end;
end;

procedure ResetNewStates;
var
  I: Integer;
begin
  for I := 0 to Length(Cells) - 1 do
    Cells[I].NewState := Cells[I].State;
end;

procedure Step;

procedure GeneralBinaryStep;
var
  C: Integer;
  Rule: Integer;

function GetRule: Integer;
const
  IndexConversion: array[0..7] of Integer = (1, 2, 4, 7, 6, 5, 3, 0);
var
  I, N, P: Integer;
begin
  N := 0;
  P := 1;
  for I := 0 to Length(Neighbors) - 1 do
  begin
    if Cells[Neighbors[IndexConversion[I]]].State = 1 then
      Inc(N, P);
    P := P * 2;
  end;
  Result := GeneralBinaryRules[N mod 16, N div 16];
end;

begin
  for C := 0 to Length(Cells) - 1 do
  begin
    GetNeighbors(C);
    Rule := GetRule;
    Cells[C].NewState := 0;
    if (Cells[C].State = 0) and ((Rule and 1) = 1) then
      Cells[C].NewState := 1
    else if (Cells[C].State = 1) and ((Rule and 2) = 2) then
      Cells[C].NewState := 1
    else if (Cells[C].State > 0) and (Cells[C].State < MaxState) then
      Cells[C].NewState := Cells[C].State + 1;
  end;
end;

procedure InteractingStatesStep;
var
  C, N: Integer;
  Counts: array[1..ArraySize] of Integer;
  I, J: Integer;
  Old, New: Integer;

label Done;

begin
  for C := 0 to Length(Cells) - 1 do
  begin
    for I := 1 to ArraySize do
      Counts[I] := 0;
    GetNeighbors(C);
    for N := 0 to Length(Neighbors) - 1 do
    begin
      if Cells[Neighbors[N]].State > 0 then
        Inc(Counts[Cells[Neighbors[N]].State]);
    end;
    New := 0;
    Old := Cells[C].State;
    if Old = 0 then
      case BirthTestOperator of
      OR_GE:
        for I := 1 to ArraySize do
          for J := 1 to ArraySize do
          if (Birth[I, J] > 0) and (Counts[J] >= Birth[I, J]) then
          begin
            New := I;
            goto Done;
          end;
      OR_EQ:
        for I := 1 to ArraySize do
          for J := 1 to ArraySize do
          if (Birth[I, J] > 0) and (Counts[J] = Birth[I, J]) then
          begin
            New := I;
            goto Done;
          end;
      AND_GE:
        for I := 1 to ArraySize do
        begin
          New := 0;
          for J := 1 to ArraySize do
            if Birth[I, J] > 0 then
            begin
              New := I;
              Break;
            end;
          for J := 1 to ArraySize do
            if (Birth[I, J] > 0) and (Counts[J] < Birth[I, J]) then
            begin
              New := 0;
              Break;
            end;
          if New = I then
            goto Done;
        end;
      AND_EQ:
        for I := 1 to ArraySize do
        begin
          New := 0;
          for J := 1 to ArraySize do
            if Birth[I, J] > 0 then
            begin
              New := I;
              Break;
            end;
          for J := 1 to ArraySize do
            if (Birth[I, J] > 0) and (Counts[J] <> Birth[I, J]) then
            begin
              New := 0;
              Break;
            end;
          if New = I then
            goto Done;
        end;
      end
    else
      case SurvivalTestOperator of
      OR_GE:
        for J := 1 to ArraySize do
        if (Survival[Old, J] > 0) and (Counts[J] >= Survival[Old, J]) then
        begin
          New := Old;
          goto Done;
        end;
      OR_EQ:
        for J := 1 to ArraySize do
          if (Survival[Old, J] > 0) and (Counts[J] = Survival[Old, J]) then
          begin
            New := Old;
            goto Done;
          end;
      AND_GE:
      begin
        for J := 1 to ArraySize do
          if Survival[Old, J] > 0 then
          begin
            New := Old;
            Break;
          end;
        for J := 1 to ArraySize do
          if (Survival[Old, J] > 0) and (Counts[J] < Survival[Old, J]) then
          begin
            New := 0;
            goto Done;
          end
      end;
      AND_EQ:
      begin
        for J := 1 to ArraySize do
          if Survival[Old, J] > 0 then
          begin
            New := Old;
            Break;
          end;
        for J := 1 to ArraySize do
          if (Survival[Old, J] > 0) and (Counts[J] <> Survival[Old, J]) then
          begin
            New := 0;
            goto Done;
          end;
      end;
    end;
    Done:  Cells[C].NewState := New;
  end;
end;

procedure WeightedGenerationsStep;
var
  C, N, I: Integer;
  Sum: Integer;

label Done;

function GetWeight(Cell: Integer; Position: Integer): Integer;
begin
  GetWeight := StateWeights[Cell] * PositionWeights[Position]
end;

begin
  for C := 0 to Length(Cells) - 1 do
  begin
    Sum := GetWeight(Cells[C].State, 4);
    GetNeighbors(C);
    for N := 0 to Length(Neighbors) - 1 do
    begin
      if N < 4 then
        I := N
      else
        I := N + 1;
      Sum := Sum + GetWeight(Cells[Neighbors[N]].State, I);
    end;
    Cells[C].NewState := 0;
    if Cells[C].State = 0 then
    begin
      for I := 1 to BirthCount do
        if Sum = BirthList[I] then
        begin
          Cells[C].NewState := 1;
          goto Done;
        end
    end;
    if Cells[C].State = 1 then
    begin
      for I := 1 to SurvivalCount do
        if Sum = SurvivalList[I] then
        begin
          Cells[C].NewState := 1;
          goto Done;
        end
    end;
    if (Cells[C].State > 0) and (Cells[C].State <> NumberOfStates - 1) then
      Cells[C].NewState := Cells[C].State + 1;
    Done:
  end;
end;

begin
  if Rule = GeneralBinaryRule then
    GeneralBinaryStep
  else if Rule = InteractingStatesRule then
    InteractingStatesStep
  else if Rule = WeightedGenerationsRule then
    WeightedGenerationsStep;
  Inc(FrameCount);
  DisplayForm.Statusbar.Panels[0].Text := IntToStr(FrameCount);
end;

procedure ProcessMouseClick(X, Y: Integer; Shift: TShiftState);
var
  I, J: Integer;

procedure AdjustSelection;
var
  T: Integer;
begin
  if SelectedBottomRight.X < SelectedTopLeft.X then
  begin
    T := SelectedBottomRight.X;
    SelectedBottomRight.X := SelectedTopLeft.X;
    SelectedTopLeft.X := T;
  end;
  if SelectedBottomRight.Y < SelectedTopLeft.Y then
  begin
    T := SelectedBottomRight.Y;
    SelectedBottomRight.Y := SelectedTopLeft.Y;
    SelectedTopLeft.Y := T;
  end;
end;

begin
  for I := 0 to Length(Cells) - 1 do
  begin
    GetGridLocation(I);
    if (X > GridLocation.X) and (X <= (GridLocation.X + GridSize)) then
      if (Y > GridLocation.Y) and (Y <= (GridLocation.Y + GridSize)) then
      begin
        if ssShift in Shift then
        begin
          GetNeighbors(I);
          for J := 0 to Length(Neighbors) - 1 do
            Cells[Neighbors[J]].NewState := 1;
        end
        else if ssCtrl in Shift then
          case Selected of
            None:
            begin
              SelectedTopLeft := GridLocation;
              SelectedBottomRight := GridLocation;
              Selected := TopLeftMarked;
            end;
            TopLeftMarked:
            begin
              SelectedBottomRight := GridLocation;
              AdjustSelection;
              X1 := (SelectedTopLeft.X - LeftBorder) div GridSize;
              Y1 := (SelectedTopLeft.Y - TopBorder) div GridSize;
              XN := (SelectedBottomRight.X - LeftBorder) div GridSize;
              YN := (SelectedBottomRight.Y - TopBorder) div GridSize;
              Selected := GroupSelected;
            end;
            GroupSelected:
              Selected := Removing;
          end
        else
          Cells[I].NewState := SelectedState;
        break;
      end;
  end;
  if Selected = Removing then
    ResetNewStates;
end;

procedure LoadPattern(FileName: string);
var
  F: TextFile;
  Line, R, S: String;

procedure AddStates(S: String);
var
  X, Y: Integer;
  N, V: Integer;

procedure AddValues(N, S: Integer);
var
  I: Integer;
begin
  if Y >= NoRows then
    Exit;
  for I := 1 to N do
  begin
    if X >= NoColumns then
      Exit;
    Cells[Y * NoColumns + X].NewState := S;
    Inc(X);
  end;
end;

begin
  X := 0;
  Y := 0;
  while S <> '' do
  begin
    N := 0;
    V := Ord(S[1]);
    while (V >= 48) and (V <= 57) do
    begin
      N := N * 10 + V - 48;
      Delete(S, 1, 1);
      V := Ord(S[1]);
    end;
    if N = 0 then
      N := 1;
    if V = 36 then
    begin
      Inc(Y, N);
      X := 0;
    end
    else
      if V = 46 then
        Inc(X, N)
      else
        AddValues(N, V - 64);
    Delete(S, 1, 1);
  end;
end;

begin
  AssignFile(F, FileName);
  Reset(F);
  S := '';
  while not Eof(F) do
  begin
    ReadLn(F, Line);
    if UpperCase(Copy(Line, 1, 20)) = '#GAME GENERAL BINARY' then
    begin
      Rule := GeneralBinaryRule;
      ReadLn(F, Line);
      while Copy(Line, 1, 6) <> '#RULE ' do
        ReadLn(F, Line);
      Delete(Line, 1, 6);
      R := Line;
    end
    else if Copy(Line, 1, 6) = '#RULE ' then
    begin
      Delete(line, 1, 6);
      if Copy(Line, 1, 18) = 'InteractingStates,' then
      begin
        Delete(line, 1, 18);
        R := Line;
        Rule := InteractingStatesRule;
      end
      else if Copy(Line, 1, 13) = 'WeightedGen3,' then
      begin
        Delete(Line, 1, 13);
        R := Line;
        Rule := WeightedGenerationsRule;
      end
      else
        R := R + Line;
    end
    else if Copy(Line, 1, 3) = '#L ' then
    begin
      Delete(Line, 1, 3);
      S := S + Line;
    end;
  end;
  CloseFile(F);
  if Rule = GeneralBinaryRule then
  begin
    if R[2] = '0' then
      R[2] := '2';
    GeneralBinaryRuleString := R
  end
  else if Rule = InteractingStatesRule then
    InteractingStatesRuleString := R
  else if Rule = WeightedGenerationsRule then
    WeightedGenerationsRuleString := R;
  AddStates(S);
end;

procedure SavePattern(FileName: string);
var
  F: TextFile;
  Line, S: String;

procedure GetStates;
var
  X, Y: Integer;
  XE: Integer;
  State, LastState: Integer;
  C: Integer;
  BlankLine: Boolean;

procedure AddItem;
begin
  if C > 1 then
    S := S + IntToStr(C);
  S := S + Char(LastState);
  if Length(S) >= 67 then
  begin
    WriteLn(F, '#L ' + S);
    S := '';
  end;
end;

begin
  C := 0;
  S := '';
  LastState := 0;
  XE := XN;
  for Y := Y1 to YN do
  begin
    BlankLine := True;
    for X := X1 to XN do
    begin
      if Cells[Y * NoColumns + X].State > 0 then
      begin
        BlankLine := False;
        break;
      end;
    end;
    if Blankline then
    begin
      if LastState = 36 then
        Inc(C)
      else
      begin
        LastState := 36;
        C := 1;
      end;
      continue;
    end
    else
      for X := XN downto X1 do
      begin
        if Cells[Y * NoColumns + X].State > 0 then
        begin
          XE := X;
          break;
        end;
      end;
    for X := X1 to XE do
    begin
      State := Cells[Y * NoColumns + X].State;
      if State = 0 then
        State := 46
      else
        State := State + 64;
      if C = 0 then
      begin
        LastState := State;
        C := 1;
      end
      else if State = LastState then
        Inc(C)
      else
      begin
        AddItem;
        LastState := State;
        C := 1;
      end;
    end;
    if LastState = 36 then
      Inc(C)
    else
    begin
      AddItem;
      LastState := 36;
      C := 1;
    end;
  end;
end;

begin
  if Rule = GeneralBinaryRule then
    S := '#RULE ' + GeneralBinaryRuleString
  else if Rule = InteractingStatesRule then
    S := '#RULE InteractingStates,' + InteractingStatesRuleString
  else if Rule = WeightedGenerationsRule then
    S := '#RULE WeightedGen3,' + WeightedGenerationsRuleString
  else
    Exit;
  Assignfile(F, FileName);
  Rewrite(F);
  WriteLn(F, '#MCell 4.20');
  if Rule = GeneralBinaryRule then
    WriteLn(F, '#GAME General Binary')
  else
    WriteLn(F, '#GAME User DLL');
  Line := Copy(S, 1, 70);
  Delete(S, 1, 70);
  WriteLn(F, Line);
  while S <> '' do
  begin
    Line := '#RULE ' + Copy(S, 1, 64);
    WriteLn(F, Line);
    Delete(S, 1, 64);
  end;
  WriteLn(F, '#BOARD ' + IntToStr(NoColumns) + 'x' + IntToStr(NoRows));
  WriteLn(F, '#WRAP 1');
  WriteLn(F, '#CCOLORS ' + IntToStr(MaxState + 1));
  GetStates;
  if S <> '' then
    Write(F, '#L ' + S);
  CloseFile(F);
end;

end.

